/*
 Chaotic 3D cat map image encryption algorithm.
 written by Yaobin Mao, July, 16, 2002
 For the detail of the proposed algorithm, refer to 
 Guanrong Chen, Yaobin Mao, Charles K. Chui, A Symmetric Image Encryption 
 Scheme Based on Chaotic 3D Cat Maps, Chaos, solitons and fractals,21(2004),
 pp.749-761.
 */
# include <stdlib.h>
# include <stdio.h>
# include <math.h>
# include <mex.h>

#define _SHOW_MESSAGE_IN_MATLAB_

struct CHEN_KEY
{
	double k_ax;
	double k_bx;
	double k_ay;
	double k_by;
	double k_az;
	double k_bz;
	double k_l;
	double k_s;
} ChenKey;

double z100x;
double z200x;
double z100y;
double z200y;
double z100z;
double z200z;
double Logistic_init;
unsigned char seed_s;

/* key generation by use of Chen's attractor */

double convert2double( unsigned char D1, unsigned char D2 )
{
	int i;
	double v = 0;
	unsigned char mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

	for ( i = 0; i < 8; i++ )
		v = v + ( D1 & mask[i] ) * 256;
	for ( i = 0; i < 8; i++ )
		v = v + ( D2 & mask[i] );
	v = 1.0/v;

	return( v );
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                Key128Bit2Double()
//  Convert user key of 128 bits (16 char) into 8 float data
//////////////////////////////////////////////////////////////////////////////// 
struct CHEN_KEY Key128Bit2Double(unsigned char * key)
{
	struct CHEN_KEY chenkey;
	
	chenkey.k_ax = convert2double( key[0], key[1] );
	chenkey.k_bx = convert2double( key[2], key[3] );
	chenkey.k_ay = convert2double( key[4], key[5] );
	chenkey.k_by = convert2double( key[6], key[7] );
	chenkey.k_az = convert2double( key[8], key[9] );
	chenkey.k_bz = convert2double( key[10], key[11] );
	chenkey.k_l = convert2double( key[12], key[13] );
	chenkey.k_s = convert2double( key[14], key[15] );

	return chenkey;
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                Chens()
//  Using Ruger-Kuta method to solute one order ordinary equation
//  for chaotic Chen's system
//  input parameters:
//               Xn, Yn, Zn --- solutions of previous step
//               a, b, c    --- control parameters
//                              when a = 35; b = 3; c = 20-28.4, the system is chaotic
//               h          --- sampling step
//   output parameters:
//               Kx, Ky, Kz --- return values, coefficients
//////////////////////////////////////////////////////////////////////////////// 

int Chens(	double * Kx,	double * Ky,	double * Kz,
			double Xn,		double Yn,		double Zn,
			double a,		double b,		double c,
			double h )
{
   Kx[0] = a * (Yn - Xn);
   Ky[0] = (c - a) * Xn - Xn * Zn + c * Yn;
   Kz[0] = Xn * Yn - b * Zn;
   Kx[1] = a * (Yn + h/2 * Ky[0] - Xn - h/2 * Kx[0]);
   Ky[1] = (c - a) * (Xn + h/2 * Kx[0]) - (Xn + h/2 * Kx[0]) * (Zn + h/2 * Kz[0]) + c * (Yn + h/2 * Ky[0]);
   Kz[1] = (Xn + h/2 * Kx[0]) * (Yn + h/2 * Ky[0]) - b * (Zn + h/2 * Kz[0]);
   Kx[2] = a * (Yn + h/2 * Ky[1] - Xn - h/2 * Kx[1]);
   Ky[2] = (c - a) * (Xn + h/2 * Kx[1]) - (Xn + h/2 * Kx[1]) * (Zn + h/2 * Kz[1]) + c * (Yn + h/2 * Ky[1]);
   Kz[2] = (Xn + h/2 * Kx[1]) * (Yn + h/2 * Ky[1]) - b * (Zn + h/2 * Kz[1]);
   Kx[3] = a * (Yn + h * Ky[2] - Xn - h * Kx[2]);
   Ky[3] = (c - a) * (Xn + h * Kx[2]) - (Xn + h * Kx[2]) * (Zn + h * Kz[2]) + c * (Yn + h * Ky[2]);
   Kz[3] = (Xn + h * Kx[2]) * (Yn + h * Ky[2]) - b * (Zn + h * Kz[2]);

   return 1;
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                 Calculate_Chens()
//  solute Chen's chaotic system, this function calls Chens()
//  input parameters: 
//               X0,Y0,Z0   ---   initial conditon
//               h          ---   sampling step
//               n          ---   calculating length, n steps
//               a,b,c      ---   control parameters
//                                when a = 35; b = 3; c = 20---28.4, the system is chaotic
//  output parameters:
//               Xn, Yn, Zn --- return value, all are vectors of nx1 in size
//  This function calls Chens()
////////////////////////////////////////////////////////////////////////////////

int Calculate_Chens(	double * Xn,	double * Yn,	double * Zn,
						double X0,		double Y0,		double Z0,
						double h,		int n,
						double a,		double b,		double c )
{
   Xn[0] = X0;
   Yn[0] = Y0;
   Zn[0] = Z0;

   double Kx[4],Ky[4],Kz[4];

   for (int i = 1; i<n; i++)
   {
    double X, Y, Z;
    X = Xn[i-1];
    Y = Yn[i-1];
    Z = Zn[i-1];
    if (Chens(Kx, Ky, Kz, X, Y, Z, a, b, c, h) <= 0) return 0;
    Xn[i] = Xn[i-1] + h/6 * (Kx[0]+2*Kx[1]+2*Kx[2]+Kx[3]);
    Yn[i] = Yn[i-1] + h/6 * (Ky[0]+2*Ky[1]+2*Ky[2]+Ky[3]);
    Zn[i] = Zn[i-1] + h/6 * (Kz[0]+2*Kz[1]+2*Kz[2]+Kz[3]);
   }
   return 1;
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                 Generate_A_Chens_Key()
//  Using Chen's system to generate cipher keys
//  
//  input parameters: 
//               chenkey  ---  parameters of generate initial value of Chens
//  output parameters:
//               z_100, z_200 --- value after 100 and 200 rounds of iteration 
//                                respectively
////////////////////////////////////////////////////////////////////////////////

#define ITERATION_OF_CHENS 200

void Generate_A_Chens_Key( double Ka, double Kb, double * z_100, double * z_200 )
{
	double *Xn = new double [ITERATION_OF_CHENS];
	double *Yn = new double [ITERATION_OF_CHENS];
	double *Zn = new double [ITERATION_OF_CHENS];
	double X0,Y0,Z0,c;

	X0 = Kb * 80 - 40;
	Y0 = Ka * 80 -40;
	Z0 = Kb * 60;
	c = Ka * 8.4 + 20;  // c belongs to [20,28.4], system chaotic
	Calculate_Chens(Xn, Yn, Zn, X0, Y0, Z0, 0.001, ITERATION_OF_CHENS, 35,3, c);
	( * z_100 ) = Zn[ITERATION_OF_CHENS/2-1];
	( * z_200 ) = Zn[ITERATION_OF_CHENS-1];
	delete Xn;
	delete Yn;
	delete Zn;
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                 Scramble_Keys()
////////////////////////////////////////////////////////////////////////////////
void Scramble_Keys( unsigned char * inKey, unsigned char * outKey)
{
	unsigned char ikey[128], tkey[128], k;
	unsigned char mask[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
	unsigned int x, y, x1, y1;
	unsigned int a = 0, b = 0;
	unsigned int fnd_table[121];

	for ( y = 0; y < 16; y++ )
		for ( x = 0; x < 8; x++ )
		{
			ikey[y*8+x] = ( mask[x] & inKey[y] )>> (7-x);
		}
	// generate find table
	for ( y = 0; y < 8; y++ )
	{
		a = a + inKey[2*y];
		b = b + inKey[2*y+1];
	}
	a = a / 8;
	b = b / 8;
	for ( x = 0; x < 11; x++ )
		for ( y = 0; y < 11; y++ )
		{
			x1 = ( (a*b+1)*x+a*y ) % 11;
			y1 = ( (b*x)+y ) % 11;
			fnd_table[x*11+y] = x1*11+y;
		}
	// circle
	tkey[127] = 1;
	for ( x = 0; x < 7; x++ )
	{
		for ( y = 0; y < 128; y++ )
			tkey[y] = ikey[y]^tkey[(128+y-1)%128];
		for ( y = x; y < x+121; y++ )
			ikey[y] = tkey[x+fnd_table[y-x]];
	}
	for ( y = 0; y < 16; y++ )
	{
		k = 0;
		for ( x = 0; x < 8; x++ )
			k = k + ikey[y*8+x]<<(7-x);
		outKey[y] = k;
	}
}

////////////////////////////////////////////////////////////////////////////////
//  function name:
//                 Generate_Keys()
//  Generate 8 useful key that will further be used in scrambling and diffusion
////////////////////////////////////////////////////////////////////////////////
void Generate_Keys( unsigned char * key )
{
	double z100, z200;
	unsigned char p_key[16];

	Scramble_Keys( key, p_key );
	ChenKey = Key128Bit2Double( p_key );
	Generate_A_Chens_Key( ChenKey.k_ax, ChenKey.k_bx, &z100x, &z200x );
	Generate_A_Chens_Key( ChenKey.k_ay, ChenKey.k_by, &z100y, &z200y );
	Generate_A_Chens_Key( ChenKey.k_az, ChenKey.k_bz, &z100z, &z200z );
	Generate_A_Chens_Key( ChenKey.k_l, ChenKey.k_s, &z100, &z200 );

	Logistic_init = z100/60;
	seed_s = (unsigned char) ( z200/60 * 255 );
}

////////////////////////////////////////////////////////////////////////////////
// function name:
//                cubic_root( int n )
//  calculate cubic root of a integer number  
////////////////////////////////////////////////////////////////////////////////
int cubic_root(int n)
{
	long a,b;

	for (int i = 0; i < 1000; i ++ )
	{
		a = i*i*i;
		b = (i+1)*(i+1)*(i+1);
		if ((a <= n) && ( n < b )) return(i);
	}
	return(0);
}

////////////////////////////////////////////////////////////////////////////////
// function name:
//                 cat3Dmap()
// make a lookup-table via 3D cat map                
////////////////////////////////////////////////////////////////////////////////
int cat3Dmap( unsigned int * table, int Width, int Height)
{
	unsigned int x, y, z, nx, ny, nz;
	unsigned long n, m, m3, begin;
	unsigned int ax, bx, ay, by, az, bz;

	n = Width*Height;
	begin = 0;
	while ( n >= 8 )
	{
		m = cubic_root( n );
		m3 = m * m * m;
		n = n - m3;
		ax = (unsigned int) (z100x / 60 * m);
		bx = (unsigned int) (z200x / 60 * m);
		ay = (unsigned int) (z100y / 60 * m);
		by = (unsigned int) (z200y / 60 * m);
		az = (unsigned int) (z100z / 60 * m);
		bz = (unsigned int) (z200z / 60 * m);
		for ( z = 0; z < m; z++ )
			for ( y = 0; y < m; y++ )
				for ( x = 0; x < m; x++ )
				{
					nx = ( (1+ax*az*by)*x + az*y + (ay+ax*az+ax*ay*az*by)*z ) % m;
					ny = ( (bz+ax*by+ax*az*by*bz)*x + (az*bz+1)*y + (ay*az+ax*ay*az*by*bz+ax*az*bz+ax*ay*by+ax)*z ) % m;
					nz = ( (ax*bx*by+by)*x + bx*y + (ax*ay*bx*by+ax*bx+ay*by+1)*z ) % m;
					table[begin + z * m * m + y * m + x] = begin + nz * m * m + ny * m + nx;
				}

		begin = begin + m3;
	}
	for ( x = begin; x < (unsigned long) Width*Height; x++ )
	{
		table[x] = x;
	}
	return( 1 );
}

////////////////////////////////////////////////////////////////////////////////
// function name:
//                 LogisticMap()
// using logistic map to generate the diffusion table
////////////////////////////////////////////////////////////////////////////////
int LogisticMap( unsigned char * d_table, int length )
{
	int i, j;
	int stop;
	double xk;

	xk = Logistic_init;
	for ( i = 0; i < 200; i++ )
		xk = 4 * xk * ( 1 - xk );
	for ( i = 0; i < length; i++ )
	{
		stop = 0;
		j = 20;
		while ( stop != 1 )
		{
			if ( xk <= 0.00001 ) xk = 0.00001;
			xk = 4 * xk * ( 1 - xk );
			j--;
			if ( ( xk <= 0.8 ) && ( xk >= 0.2 ) && ( j<= 0) ) stop = 1;
		}
		d_table[i] = (unsigned char) (( xk-0.2 ) / 0.6 * 255);
	}
	return( 1 );
}

////////////////////////////////////////////////////////////////////////////////
// function name:
//                 Cat3DEncryption()
////////////////////////////////////////////////////////////////////////////////
int Cat3DEncryption( unsigned char * pImage, 
					 unsigned char * cImage, 
					 int Width, 
					 int Height, 
					 int round )
{
	int i, r;
	long total;
	unsigned int * table;
	unsigned char * d_table, Ik;

	total = Width * Height;
	table = new unsigned int [total]; 
	d_table = new unsigned char [total]; 
	cat3Dmap( table, Width, Height);	// generate permutation table
	LogisticMap( d_table, total );		// generate diffusion table
	for ( r = 0; r < round; r++ )
	{
		/* permute data through 3D cat map */
		for ( i = 0; i < total; i++ )
			cImage[table[i]] = pImage[i];
		/* diffusion data */
		Ik = seed_s;    // initial seed
		for ( i = 0; i < total; i++ )
		{
			cImage[i] = d_table[i] ^ ( ( cImage[i] + d_table[i] ) % 256 ) ^ Ik;
			Ik = cImage[i];
		}
		/* up-down the data */
		for ( i = 0; i < total; i++ ) pImage[i] = cImage[total-i-1];
	}
	delete table;
	delete d_table;

	return( 1 );
}

////////////////////////////////////////////////////////////////////////////////
// function name:
//                 Cat3DDecryption()
////////////////////////////////////////////////////////////////////////////////
int Cat3DDecryption( unsigned char * cImage, 
					 unsigned char * dImage, 
					 int Width, 
					 int Height, 
					 int round )
{
	int i, r;
	long total;
	unsigned int * table;
	unsigned char * d_table, Ik;

	total = Width * Height;
	table = new unsigned int [total]; 
	d_table = new unsigned char [total]; 
	cat3Dmap( table, Width, Height);	// generate permutation table
	LogisticMap( d_table, total );		// generate diffusion table
	for ( r = 0; r < round; r++ )
	{
		/* de-diffusion data */
		Ik = seed_s;    // initial seed
		for ( i = 0; i < total; i++ )
		{
			dImage[i] = ( (d_table[i] ^ cImage[i] ^ Ik) + 256 - d_table[i] ) % 256;
			Ik = cImage[i];
		}
		/* de-permute data through 3D cat map */
		for ( i = 0; i < total; i++ )
			cImage[i] = dImage[table[i]];
		/* up-down the data */
		if ( r < round-1 )
		{
			for ( i = 0; i < total; i++ ) dImage[i] = cImage[total-i-1];
			for ( i = 0; i < total; i++ ) cImage[i] = dImage[i];			
		}
		else
			for ( i = 0; i < total; i++ ) dImage[i] = cImage[i];
	}
	delete table;
	delete d_table;

	return( 1 );
}

/* interface with Matlab */

void mexFunction( int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[] )
{
    int H1, W1, C1;
    int number_of_dims;
    const int * dim_array;
    int dims[3];
    double * data1, * data2, * key;
    int x, y, z;
    long length;
    int m,n;
    unsigned int round;
	unsigned char keyChar[17];
    
    /* check the validity of input and output parameters */
    if ( nrhs < 3 )
    {
    	mexErrMsgTxt( "At least three input arguments is required." );
    	return;
    }
    else if (nlhs > 1 )
    {
    	mexErrMsgTxt( "Too many output arguments." );
    	return;
    }
    
    /* Get the number of dimensions in the input argument. */
    number_of_dims = mxGetNumberOfDimensions(prhs[0]);
    dim_array = mxGetDimensions(prhs[0]);
    if ( number_of_dims == 3 )
    {
		H1 = dim_array[0];            /* three dimension         */
		W1 = dim_array[1];
		C1 = dim_array[2];
    }
    else
    {
		H1 = dim_array[0];             /* two dimension          */
		W1 = dim_array[1];
		C1 = 1;
    }
	if ( C1 != 1 ) 
    {
		mexErrMsgTxt("Currently only gray-scale images are supported.\n");
		return;
    }
    if (!mxIsDouble(prhs[0])) 
    {
		mexErrMsgTxt("Elements in matrix should be double type.");
		return;
    }
    data1 = mxGetPr(prhs[0]);
    length = H1*W1*C1;
    
    /* check the second parameter to be sure it is of double */
    m = mxGetM(prhs[1]);
    n = mxGetN(prhs[1]);
    if (!mxIsDouble(prhs[1])) 
    {
		mexErrMsgTxt("Elements in second matrix should be of double type.");
		return;
    }
    key = mxGetPr(prhs[1]);
    round = (unsigned int)(key[0]);
    
    /* check the third parameter to be sure it is of double */
    m = mxGetM(prhs[2]);
    n = mxGetN(prhs[2]);
    if (!mxIsDouble(prhs[2])) 
    {
		mexErrMsgTxt("Elements in second matrix should be double type.");
		return;
    }
    if ( m*n != 16 ) 
    {
		mexErrMsgTxt("Key should be 16 bits of length.");
		return;
    }	
    key = mxGetPr(prhs[2]);
    for(y=0; y<m; y++)
    {
		for(x=0; x<n; x++)
		{
			keyChar[y*n+x] = (unsigned char)key[y + x * m];	  // set key;
		}
    }
    
    /* create output matrix */
    dims[0] = H1; dims[1] = W1; dims[2] = C1;
    plhs[0] = mxCreateNumericArray( 3, dims, mxDOUBLE_CLASS, mxREAL );
    data2 = mxGetPr(plhs[0]);

    /* allocate memory */
    unsigned char * pData = new unsigned char[W1 * H1 * C1];

    for ( z = 0; z < C1; z++ )
	for ( x = 0; x < W1; x++ )
	    for ( y = 0; y < H1; y++ )
	    {   
	    	pData[y + x * H1 + z * W1 * H1] = (unsigned char) data1[y + x * H1 + z * W1 * H1];
	    }
    
    unsigned char * outData = new unsigned char [W1 * H1 * C1];

    /* encryption */
	Generate_Keys( keyChar );     // generate key;
	Cat3DEncryption( pData, outData, W1, H1, round ); // perform encryption
    
    /* output result */
    for ( z = 0; z < C1; z++ )
	for ( x = 0; x < W1; x++ )
	    for ( y = 0; y < H1; y++ )
	    {   
	    	data2[y + x * H1 + z * W1 * H1] = outData[y + x * H1 + z * W1 * H1];
	    }

   delete []pData;
   delete []outData;
   pData = NULL;
   outData = NULL;
	        
   return;
}
