/*********************************************************************************
 Chaotic 3D baker map image encryption algorithm.
 written by Yaobin Mao, Auguest, 18, 2002
 revised on May, 16, 2006
 For the detail of the proposed algorithm, please refer to 

 Yaobin Mao, Guanrong Chen, Shiguo Lian. A Novel Fast Image Encryption Scheme 
 Based on 3D Chaotic Baker Maps. International Journal of Bifurcation and Chaos, 
 Vol. 14, No. 10 (2004) 3613-3624.
 *********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "mex.h"

typedef struct Factor_Element
{
	int Factor;
	struct Factor_Element * next;
} FACTOR_ELEMENT;

typedef struct Factor_List
{
	FACTOR_ELEMENT * list_header;
	int element_numbers;
} FACTOR_LIST;

/* List operation functions */
FACTOR_LIST * creat_factor_list(void)
{
	FACTOR_LIST * list;
	list = new FACTOR_LIST;
	if (list == NULL ) return NULL;
	list->list_header = NULL;
	list->element_numbers = 0;

	return(list);
}

FACTOR_LIST * add_one_element(FACTOR_LIST * list, int data)
{
	FACTOR_ELEMENT * p, * m, * parent;
	p = new FACTOR_ELEMENT;
	if (p == NULL) return(list);
	p->Factor = data;
	p->next = NULL;
	if (list!=NULL)
	{
		if (list->list_header == NULL)
		{
			list->list_header = p;
		}
		else
		{
			m = list->list_header;
			while(m!=NULL) 
			{
				parent = m;
				m = m->next;
			}
			parent->next = p;
		}
		list->element_numbers++;
	}
	return(list);
}

int get_all_factors(FACTOR_LIST * list, int * data)
{
	FACTOR_ELEMENT * m;
	int counter = 0;

	if (list!=NULL)
	{
		m = list->list_header;
		while( m != NULL )
		{
			data[counter] = m->Factor;
			counter++;
			m = m->next;
		}
	}
	return(counter);
}

FACTOR_LIST * delete_all_element(FACTOR_LIST * list)
{
	FACTOR_ELEMENT * m, * t;

	if (list!=NULL)
	{
		m = list->list_header;
		while( m != NULL )
		{
			t = m;
			m = m->next;
			delete t;
		}
		delete list;
	}
	return(NULL);
}

int get_factors(int n, FACTOR_LIST * list)
{
	int i, mid = n;

	while ( mid!=1 )
	{
		if (mid % 2 == 0) /* only even number */
		{
			list = add_one_element( list, 2 );
			mid = mid/2;
		}
		for (i = 3; i <= mid; i = i+2) /* all primes except 2 are odd numbers */
		{
			if ( mid % i == 0 )
			{
				list = add_one_element( list, i );
				mid = mid/i;
				break;
			}
		}
	}

	return(list->element_numbers);
}

/* permute a vector with random number
 A           --- resource array
 length_of_A --- length of vector A
 seed        --- random seed
 times       --- times of permutation
 Example:
 Permute_N_Point([1,2,3,4,5], 80, 3);
 result:[4 1 3 5 2]
 */
void Permute_N_Point( int * A, int length_of_A, unsigned int seed, int times)
{
	int t, i, Pos, tmp;
	unsigned int seed1;

	seed1 = seed;
	for ( t = 0; t < times; t++)
	{
		srand(seed1);
		seed1 = rand();
		for ( i = 0; i < length_of_A; i++)
		{
			Pos = rand() * i/RAND_MAX;
			tmp = A[i];
			A[i] = A[Pos];
			A[Pos] = tmp;
		}
	}

	return;
}

/* from 2D to 3D */
/* get WxLxH from a WidthxHeight image */
void Get_2D_to_3D_parameters( int width, int height,             /* original 2D image's width and height*/
							  int * W, int * L, int * H,         /* decomposed W(idth),L(ength) and H(eight) of a cube */
							  int seed_key, int shuffle_times)   /* two keys represent seed and shuffle times respectively */
{
	int * data;
	FACTOR_LIST * lst;

	int n, tmp;
	n = width * height;
	lst = creat_factor_list();
	int num = get_factors( n, lst);
	data = new int [num];
	num = get_all_factors( lst, data );
	lst = delete_all_element(lst);
	Permute_N_Point( data, num, seed_key, shuffle_times);
	
	* W = 1;
	* L = 1;
	* H = 1;
	for (int i = 0; i < num; i = i+3)
	{
		* W = (* W) * data[i];
		if ( (i+1) < num ) * L = (* L) * data[i+1];
		if ( (i+2) < num ) * H = (* H) * data[i+2];
	}
	delete data;
	if ( (*W) < (*H) )
	{
		tmp = *W;
		*W = *H;
		*H = tmp;
	}
	if ( (*L) < (*H) )
	{
		tmp = *L;
		*L = *H;
		*H = tmp;
	}

	return;
}

#define MAX_INTERVAL 30

void key_generation( char * key,  int length,      int width,   int colors,
		     int * key_l, int * num_key_l, int * key_w, int * num_key_w,
		     double * logistic_key, int * XOR_key, int * seed_key, int * shuffle_times)
{
	double Lo1;
	int remain, key_counter,tmp, Interval, i;
	char tp;

	/* generate key_l */
	Lo1 = ((key[0]*256.0 + key[1])*256.0 + key[2])/16777216.0;
	if ( (Lo1 <= 0) || (Lo1 >=1) || (Lo1 == 0.5) ) Lo1 = 0.5001; /* three bad point 0,1 and 0.5 */
	for (i = 0; i < 50 + key[0]; i++) Lo1 = 4.0*Lo1*(1.0-Lo1);
	remain = length;
	Interval = remain > MAX_INTERVAL ? MAX_INTERVAL : remain;
	key_counter = 0;
	do 
	{
		do
		{
			Lo1 = 4.0*Lo1*(1.0-Lo1);
		}while( (Lo1<0.2) || (Lo1>0.8) );
		Interval = remain > MAX_INTERVAL ? MAX_INTERVAL : remain;		
		tmp = (int)(Interval*(Lo1-0.2)/0.6);
		if ( tmp > 0 )
		{
			key_l[key_counter] = tmp;
			remain = remain-key_l[key_counter++];
		}
	} while( remain>1 );
	if (remain == 1) key_l[key_counter++] = remain;
	* num_key_l = key_counter;
	/* generate key_w */
	Lo1 = ((key[3]*256.0 + key[4])*256.0 + key[5])/16777216.0;
	if ( (Lo1 <= 0) || (Lo1 >=1) || (Lo1 == 0.5) ) Lo1 = 0.5001; /* three bad point 0,1 and 0.5 */
	for (i = 0; i < 50 + key[3]; i++) Lo1 = 4.0*Lo1*(1.0-Lo1);	
	remain = width;
	Interval = remain > MAX_INTERVAL ? MAX_INTERVAL : remain;
	key_counter = 0;
	do 
	{
		do
		{
			Lo1 = 4.0*Lo1*(1.0-Lo1);
		}while( (Lo1<0.2) || (Lo1>0.8) );
		Interval = remain > MAX_INTERVAL ? MAX_INTERVAL : remain;		
		tmp = (int)(Interval*(Lo1-0.2)/0.6);
		if ( tmp > 0 )
		{
			key_w[key_counter] = tmp;
			remain = remain-key_w[key_counter++];
		}
	} while( remain>1 );
	if (remain == 1) key_w[key_counter++] = remain;
	* num_key_w = key_counter;
	/* generate logistic_key */
	Lo1 = ((key[6]*256.0 + key[7])*256.0 + key[8])/16777216.0;
	if ( (Lo1 <= 0) || (Lo1 >=1) || (Lo1 == 0.5) ) Lo1 = 0.5001; /* three bad point 0,1 and 0.5 */
	for (i = 0; i < 100 + key[6]; i++) Lo1 = 4.0*Lo1*(1.0-Lo1);
	* logistic_key = Lo1;
	/* generate XOR key */
	switch(colors)
	{
		case 2:
			tp = key[9] ^ key[10] ^ key[11];
			tp = (tp&0x0f) ^ ((tp&0xf0)>>4);
			tp = (tp&0x03) ^ ((tp&0x0c)>>2);
			* XOR_key = (tp&0x01) ^ ((tp&0x02)>>1);
			break;
		case 16:
			* XOR_key = (key[9]&0x0f) ^ ((key[9]&0xf0)>>4) ^ (key[10]&0x0f) ^ ((key[10]&0xf0)>>4) ^ (key[11]&0x0f) ^ ((key[11]&0xf0)>>4);
			break;
		case 256:
			* XOR_key = key[9] ^ key[10] ^ key[11];
			break;
		case 24:
		default:
			* XOR_key = ((int)(key[9])*256 + key[10])*256 + key[11];
			break;
	}
	/* generate seed_key */
	* seed_key = (int)(key[12])*256+key[13];
	/* generate shuffle_times */
	* shuffle_times = (int)key[14]+key[15];
	if ((*shuffle_times) < 5) * shuffle_times = 5;
	/* permute key_l and key_w */
	Permute_N_Point( key_l, * num_key_l, (unsigned int) * seed_key, * shuffle_times);
	Permute_N_Point( key_w, * num_key_w, (unsigned int) * seed_key, * shuffle_times);
	return;
}

# define FORWARD_BAKER    1
# define INVERT_BAKER     0

int * Baker_3D( int * in_data,                      /* input data */
		int L,         int W,       int H,  /* height, width and length of the cube */
		int * key_l,   int * key_w,         /* key vectors in length and width */
		int num_key_l, int num_key_w,       /* key vector length of l and w */
		double Logistic_key,                /* logistic map */
		int XOR_key,                        /* Xor key */
		int colors,                         /* can be 2,16,256,24bits */
		int direction )                     /* 1 --- forward, 0 --- invert */
{
	int * buf, i, j, l, counter;
	int in, jn, mi, nj;
	long t1,t2,t;
	double Lo1;
	int XOR_N,Ik_1;
	int tt;

	buf = new int [L*W*H];

	if (direction == FORWARD_BAKER) 
	{
		/* encipher */
		/* permuting */
		counter = 0;
		nj = 0;
		for (jn = 0; jn < num_key_l; jn++)
		{
			mi = 0;
			for (in = 0; in < num_key_w; in++)
			{
				for (l = 0; l < H; l++)
				{
					t1 = l*L*W;
					for (j = 0; j < key_l[jn]; j++)
					{
						t2 = (j+nj)*W;
						for (i = 0; i < key_w[in]; i++)
						{
							t = t1+t2+(i+mi);
							buf[counter++] = in_data[t]; 
						}
					}
				}
				mi = mi + key_w[in];
			}
			nj = nj + key_l[jn];		
		}
		/* diffusion */
		Lo1 = Logistic_key;
		Ik_1 = XOR_key;
		if ( (Lo1 <= 0) || (Lo1 >=1) || (Lo1 == 0.5) ) Lo1 = 0.5001; /* three bad point 0,1 and 0.5 */
		for (i = 0; i < W*L*H; i++)
		{
			do 
			{
				Lo1 = 4.0*Lo1*(1.0-Lo1);
			}while( (Lo1<0.2) || (Lo1>0.8) );

			switch(colors)
			{
			case 2:
				XOR_N = (int)((Lo1 - 0.2)*1/0.6);
				in_data[i] = buf[i] ^ Ik_1 ^ XOR_N;
				Ik_1 = in_data[i];
				break;
			case 16:
				XOR_N = (int)((Lo1 - 0.2)*15/0.6);
				in_data[i] = buf[i] ^ Ik_1 ^ XOR_N;
				Ik_1 = in_data[i];
				break;
			case 256:
				XOR_N = (int)((Lo1 - 0.2)*255/0.6);
				//in_data[i] = buf[i] ^ Ik_1 ^ XOR_N;
				// following two sentence is to diffuse data by a small perturbation
				tt = ( buf[i] + XOR_N ) % 256;
				in_data[i] = XOR_N ^ tt ^ Ik_1;
				Ik_1 = in_data[i];
				break;
			case 24:
			default:
				XOR_N = (int)((Lo1 - 0.2)*16777215/0.6);
				in_data[i] = buf[i] ^ Ik_1 ^ XOR_N;	
				Ik_1 = in_data[i];
				break;
			}
		}
	}
	else
	{
		/* decipher */
		/* invert diffusion */
		Lo1 = Logistic_key;
		Ik_1 = XOR_key;
		if ( (Lo1 <= 0) || (Lo1 >=1) || (Lo1 == 0.5) ) Lo1 = 0.5001; /* three bad point 0,1 and 0.5 */
		for (i = 0; i < W*L*H; i++)
		{
			do 
			{
				Lo1 = 4.0*Lo1*(1.0-Lo1);
			}while( (Lo1<0.2) || (Lo1>0.8) );

			switch(colors)
			{
			case 2:
				XOR_N = (int)((Lo1 - 0.2)*1/0.6);
				buf[i] = Ik_1 ^ in_data[i] ^ XOR_N;
				Ik_1 = in_data[i];
				break;
			case 16:
				XOR_N = (int)((Lo1 - 0.2)*15/0.6);
				buf[i] = Ik_1 ^ in_data[i] ^ XOR_N;
				Ik_1 = in_data[i];
				break;
			case 256:
				XOR_N = (int)((Lo1 - 0.2)*255/0.6);
				//buf[i] = Ik_1 ^ in_data[i] ^ XOR_N;
				// following two sentences is to un-diffuse data
				tt = Ik_1 ^ in_data[i] ^ XOR_N;
				buf[i] = ( tt  + (256-XOR_N) ) % 256 ;
				Ik_1 = in_data[i];
				break;
			case 24:
			default:
				XOR_N = (int)((Lo1 - 0.2)*16777215/0.6);
				buf[i] = Ik_1 ^ in_data[i] ^ XOR_N;
				Ik_1 = in_data[i];
				break;
			}
		}
		/* invert permuting */
		counter = 0;
		nj = 0;
		for (jn = 0; jn < num_key_l; jn++)
		{
			mi = 0;
			for (in = 0; in < num_key_w; in++)
			{
				for (l = 0; l < H; l++)
				{
					t1 = l*L*W;
					for (j = 0; j < key_l[jn]; j++)
					{
						t2 = (j+nj)*W;
						for (i = 0; i < key_w[in]; i++)
						{
							t = t1+t2+(i+mi);
							in_data[t] = buf[counter++]; 
						}
					}
				}
				mi = mi + key_w[in];
			}
			nj = nj + key_l[jn];		
		}
	}

	delete buf;
	return in_data;
}

int * Bakermap_3D_Encipher_Image( int *pData, char * key, int iHeight, int iWidth, int iColorBits, int iRound )
{
	int colors;
	int * key_l, * key_w;
	int num_key_l, num_key_w;
	double logistic_key;
	int XOR_key, seed_key, shuffle_times;
	int W,L,H;
	char temp;
	int i,j;

	seed_key = (int)(key[12])*256+key[13];
	shuffle_times = (int)key[14]+key[15];
	Get_2D_to_3D_parameters(iWidth, iHeight, &W, &L, &H, seed_key, shuffle_times);
	switch( iColorBits )
	{
	case 1:
		colors = 2;
		break;
	case 4:
		colors = 16;
		break;
	case 8:
		colors = 256;
		break;
	case 24:
		colors = 24;
	default:
		break;
	}
	key_l = new int [L];
	key_w = new int [W];
	for (i = 0; i < iRound; i++ )
	{
		key_generation( key, L, W, colors, key_l, &num_key_l, key_w, &num_key_w, 
		                &logistic_key, &XOR_key, &seed_key, &shuffle_times);
		pData = Baker_3D( pData, L, W, H, key_l, key_w, num_key_l, num_key_w,
			          logistic_key, XOR_key, colors, FORWARD_BAKER );
		for (j = 0; j < 8; j++)
		{
			temp = key[j];
			key[j] = key[15-j];
			key[15-j] = temp;
		}
	}
	delete key_l;
	delete key_w;
	return(pData);
}

int * Bakermap_3D_Decipher_Image( int *pData, char * key, int iHeight, int iWidth, int iColorBits, int iRound )
{
	int colors;
	int * key_l, * key_w;
	int num_key_l, num_key_w;
	double logistic_key;
	int XOR_key, seed_key, shuffle_times;
	int W,L,H;
	char temp;
	int i,j;

	seed_key = (int)(key[12])*256+key[13];
	shuffle_times = (int)key[14]+key[15];
	Get_2D_to_3D_parameters(iWidth, iHeight, &W, &L, &H, seed_key, shuffle_times);
	switch( iColorBits )
	{
	case 1:
		colors = 2;
		break;
	case 4:
		colors = 16;
		break;
	case 8:
		colors = 256;
		break;
	case 24:
		colors = 24;
	default:
		break;
	}
	key_l = new int [L];
	key_w = new int [W];
	if ( iRound % 2 != 0 )     /* if iRound is odd */
	{
		for (j = 0; j < 8; j++)
		{
			temp = key[j];
			key[j] = key[15-j];
			key[15-j] = temp;
		}
	}
	for (i = 0; i < iRound; i++ )
	{
		for (j = 0; j < 8; j++)
		{
			temp = key[j];
			key[j] = key[15-j];
			key[15-j] = temp;
		}
		key_generation( key, L, W, colors, key_l, &num_key_l, key_w, &num_key_w, 
		                &logistic_key, &XOR_key, &seed_key, &shuffle_times);
		pData = Baker_3D( pData, L, W, H, key_l, key_w, num_key_l, num_key_w,
		  	          logistic_key, XOR_key, colors, INVERT_BAKER );
	}
	delete key_l;
	delete key_w;
	return(pData);
}

/* interface with Matlab */
void mexFunction( int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[] )
{
    int H1, W1, C1;
    int number_of_dims;
    const int * dim_array;
    int dims[3];
    double * data1, * data2, * key;
    int x, y;
    long length;
    int m,n;
    unsigned int round;
    char keyChar[17];
    
    /* check the validity of input and output parameters */
    if ( nrhs < 3 )
    {
    	mexErrMsgTxt( "At least three input arguments is required." );
    	return;
    }
    else if (nlhs > 1 )
    {
    	mexErrMsgTxt( "Too many output arguments." );
    	return;
    }
    
    /* The first parameter is image array */
    /* Get the number of dimensions in the input argument. */
    number_of_dims = mxGetNumberOfDimensions(prhs[0]);
    dim_array = mxGetDimensions(prhs[0]);
    if ( number_of_dims == 3 )
    {
		H1 = dim_array[0];            /* three dimension         */
		W1 = dim_array[1];
		C1 = dim_array[2];
    }
    else
    {
		H1 = dim_array[0];             /* two dimension          */
		W1 = dim_array[1];
		C1 = 1;
    }
    if ( !mxIsDouble(prhs[0]) ) 
    {
		mexErrMsgTxt("Elements in matrix should be of double type.");
		return;
    }
    data1 = mxGetPr(prhs[0]);
    length = H1*W1*C1;
    
    /* the second parameter is rounds to performed */
    /* check the second parameter to be sure it is of double */
    m = mxGetM(prhs[1]);
    n = mxGetN(prhs[1]);
    if (!mxIsDouble(prhs[1])) 
    {
		mexErrMsgTxt("Elements in second matrix should be of double type.");
		return;
    }
    key = mxGetPr(prhs[1]);
    round = (unsigned int)(key[0]);
    
    /* the third parameter is 16-char-long key */
    /* check the third parameter to be sure it is of double */
    m = mxGetM(prhs[2]);
    n = mxGetN(prhs[2]);
    if (!mxIsDouble(prhs[2])) 
    {
		mexErrMsgTxt("Elements in second matrix should be double type.");
		return;
    }
    if ( m*n != 16 ) 
    {
		mexErrMsgTxt("Key should be 16 bits of length.");
		return;
    }	
    key = mxGetPr(prhs[2]);
    for(y=0; y<m; y++)
    {
		for(x=0; x<n; x++)
		{
			keyChar[y*n+x] = (unsigned char)key[y + x * m];	  // set key;
		}
    }
    
    /* create output matrix */
    dims[0] = H1; dims[1] = W1; dims[2] = C1;
    plhs[0] = mxCreateNumericArray( 3, dims, mxDOUBLE_CLASS, mxREAL );
    data2 = mxGetPr(plhs[0]);

    /* allocate memory */
    int * pData = new int [W1 * H1];

    if ( C1 == 1 )
    {
    	for ( x = 0; x < W1; x++ )
	    for ( y = 0; y < H1; y++ )
	    {
	    	pData[y + x * H1] = (unsigned char) data1[y + x * H1];
	    }	    	
    }
    else
    {
	for ( x = 0; x < W1; x++ )
	    for ( y = 0; y < H1; y++ )
	    {   
	    	pData[y + x * H1] = (unsigned char) data1[y + x * H1 + 2 * W1 * H1] * 256 * 256 
	    			  + (unsigned char) data1[y + x * H1 + W1 * H1] * 256 
	    			  + (unsigned char) data1[y + x * H1];
	    }
    }

    /* perform encryption */
    if ( C1 == 1 )
	pData = Bakermap_3D_Decipher_Image( pData, keyChar, H1, W1, 8, round );
    else
	pData = Bakermap_3D_Decipher_Image( pData, keyChar, H1, W1, 24, round ); 
    
    /* output result */
    for ( x = 0; x < W1; x++ )
	for ( y = 0; y < H1; y++ )
	{   
		data2[y + x * H1 + 2 * W1 * H1] = pData[y + x * H1]/256/256;
		data2[y + x * H1 + 1 * W1 * H1] = (pData[y + x * H1]/256)%256;
		data2[y + x * H1] = (pData[y + x * H1]%256)%256;
	}

   delete []pData;
   pData = NULL;
	        
   return;
}
